{-
Print library names in the top level of /usr/lib. To run:

    stack runghc --package terminal-size example/Libnames.hs

Zsh equivalent:

    #!/usr/bin/env zsh
    setopt histsubstpattern
    libnames=( /usr/lib/lib*(.:t:s/[.-]*//:s/#lib//) )
    print -c ${(u)libnames}

stack runghc --package terminal-size example/Libnames.hs
-}

import Hexpand

import Control.Monad
import Data.List (elemIndex, group, sort)
import Data.Maybe (fromMaybe)
import qualified Data.IntMap.Strict as IM

import System.Console.Terminal.Size as Terminal

main :: IO ()
main = do
    libnames <- getLibnames
    prettyPrint libnames

dedup :: (Ord a) => [a] -> [a]
dedup = map head . group . sort

getLibnames :: IO [String]
getLibnames = do
    libFilenames <- liftM2 fromMaybe (return []) (glob "/usr/lib/lib*(-.:t)")
    return $ dedup (map transform libFilenames)
  where
    -- Drop the "lib" prefix as well as version number (if nay) and extension
    transform f = drop 3 (dropSuffix f)
    -- Dropped suffix start from the first occurrence of dash (-1.0) or dot
    -- (.1.0, .a, .dylib, etc.)
    dropSuffix f = take (min dashPos dotPos) f
      where
        -- Character index in string (first occurrence, or the length of the
        -- string if nonexistent)
        charPos c s = fromMaybe (length f) (elemIndex c s)
        dashPos     = charPos '-' f
        dotPos      = charPos '.' f

prettyPrint :: [String] -> IO ()
prettyPrint xs = do
    maybeWindow <- Terminal.size
    case maybeWindow of
        Just (Window _ w) -> prettyPrintInWidth w xs
        -- Terminal width not available, print one entry per line
        otherwise -> mapM_ putStrLn xs

prettyPrintInWidth :: Int -> [String] -> IO () -- String
prettyPrintInWidth width xs =
    mapM_ putStrLn rows
  where
    nelems = length xs
    longestElemLen = maximum $ map length xs
    colWidth = longestElemLen + 2
    ncols = width `quot` colWidth
    nrows = ((nelems - 1) `quot` ncols) + 1
    -- Pair each padded element to its row index (from 0)
    rowElemPairs = map (\n -> (n `rem` nrows, pad (xs !! n))) [0..nelems - 1]
    pad elem = elem ++ replicate (colWidth - length elem) ' '
    rows = IM.elems $ IM.fromListWith append rowElemPairs
    append elem sofar = sofar ++ elem
