{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Hexpand
    ( glob
    ) where

import Control.Applicative (many)
import Control.Exception (catch)
import Control.Monad
import Data.Attoparsec.ByteString.Char8
import Data.ByteString.Char8 (pack, unpack)
import Data.Char
import Data.List (isPrefixOf, sort, sortBy, stripPrefix)
import System.Directory
import System.FilePath
import System.IO
import System.Posix.Files
import qualified Data.Map.Strict as M


-- |Representation of a complete pattern (with qualifiers and modifiers).
--
-- The first Bool indicates whether the pattern is absolute (starts with /) or
-- relative (otherwise).
data Pattern = Pattern Bool [PatternSegment] [Qualifier] [Modifier]
    deriving (Eq, Ord, Show)

type Qualifier = String

type Modifier = String

type PatternSegment = [PatternUnit]

-- TODO: Extended glob operators (^x, x~y, x#, and x##)
-- TODO: Support for . and ..
-- TODO: Unicode support
-- TODO: Special handling of dotfiles
data PatternUnit = Literal String
                 | Star
                 | DoubleStar
                 | TripleStar
                 | QuestionMark
                 | CharClass { includedChars :: [Char] }
                 | InvertedCharClass { excludedChars :: [Char] }
                 | NumberRange { lower :: Integer
                               , upper :: Integer }
                 | Group PatternSegment
                 | Alternation [PatternSegment]
    deriving (Eq, Ord, Show)


glob :: String -> IO (Maybe [FilePath])
glob s = case extractQualifiersAndModifiers s of
    Nothing -> qmError
    Just (s', quals, modifiers) -> do
        -- Perform basic tilde expansion for current user
        s'' <- if s' == "~"
               then getHomeDirectory
               else if "~/" `isPrefixOf` s'
               then do home <- getHomeDirectory
                       return (home </> (drop 2 s'))
               else return s'
        case parseOnly pattern (pack s'') of
            Left err -> patternError
            Right (absolute, ps) ->
                liftM Just (glob' (Pattern absolute ps quals modifiers))
  where
    -- TODO: Refactor: Return an error instead of printing to stderr
    patternError = do
        hPutStrLn stderr "Error: Bad pattern"
        return Nothing
    qmError = do
        hPutStrLn stderr "Error: Bad qualifier or modifier"
        return Nothing

glob' :: Pattern -> IO [FilePath]
glob' (Pattern absolute ps qs ms) = do
    root <- if absolute
            then return "/"
            else getCurrentDirectory
    fs <- matchBuilder root ps
    let fs' = if absolute
              then map (\f -> "/" ++ f) fs
              else fs
    let fs'' = sort fs'
    fs''' <- applyQualifiers fs'' qs
    fs'''' <- applyModifiers fs''' ms
    return fs''''

-- |Sort by Keys with a key generator. The key for each element is only
-- generated once.
sortByKey :: (Ord a, Ord b) => (a -> b) -> [a] -> [a]
sortByKey keyfun xs = sortBy cmp xs
  where
    elemKeyMap = M.fromList $ map (\x -> (x, keyfun x)) xs
    cmp x y = case M.lookup x elemKeyMap of
        Nothing -> EQ
        Just kx -> case M.lookup y elemKeyMap of
            Nothing -> EQ
            Just ky -> compare kx ky

-- |Sort by keys with a monadic key generator. The key for each element is only
-- generated once.
sortByKeyM :: (Ord a, Ord b, Monad m) => (a -> m b) -> [a] -> m [a]
sortByKeyM keyfun xs = do
    let elemToKeyElemPair x = do k <- keyfun x
                                 return (x, k)
    elemKeyMap <- liftM M.fromList (mapM elemToKeyElemPair xs)
    let cmp x y = case M.lookup x elemKeyMap of
            Nothing -> EQ
            Just kx -> case M.lookup y elemKeyMap of
                Nothing -> EQ
                Just ky -> compare kx ky
    return $ sortBy cmp xs

-- TODO: Handle multiple order qualifiers (later order qualifiers should be
-- used as tie breaker, not overriding)
applyQualifiers :: [FilePath] -> [Qualifier] -> IO [FilePath]
applyQualifiers fs qs = do
    foldM apply fs qs
  where
    -- Apply qualifier to file list
    -- apply :: [FilePath] -> Qualifier -> IO [FilePath]
    apply fs q = case q of
        -- Directories only
        "/" -> filterM (fromSymStatus isDirectory) fs
        -- Regular files only
        "." -> filterM (fromSymStatus isRegularFile) fs
        -- Symlinks only
        "@" -> filterM (fromSymStatus isSymbolicLink) fs
        -- Dir or dir symlinks only
        "-/" -> filterM (fromStatus isDirectory) fs
        -- File or file symlinks only
        "-." -> filterM (fromStatus isRegularFile) fs
        -- Sort by name
        "on" -> return $ sort fs
        -- Sort by file size
        "oL" -> sortByKeyM (fromStatus fileSize) fs
        -- Sort by link count
        "ol" -> sortByKeyM (fromSymStatus linkCount) fs
        -- Sort by atime (latest first)
        "oa" -> reverseM $ sortByKeyM (fromStatus accessTimeHiRes) fs
        -- Sort by mtime (latest first)
        "om" -> reverseM $ sortByKeyM (fromStatus modificationTimeHiRes) fs
        -- Sort by ctime (latest first)
        "oc" -> reverseM $ sortByKeyM (fromStatus statusChangeTimeHiRes) fs
        otherwise -> return fs
    -- getFileStatus on the file, then run the given function (promoted to monad)
    -- fromStatus :: (FileStatus -> a) -> FilePath -> IO a
    fromStatus func f = liftM func (getFileStatus f)
    -- Same as fromStatus, except getSymbolicLinkStatus is used in place of getFileStatus
    fromSymStatus func f = liftM func (getSymbolicLinkStatus f)
    reverseM = liftM reverse

-- TODO: Handle ordering of modifiers (e.g. :h:r could be different from :r:h)
applyModifiers :: [FilePath] -> [Modifier] -> IO [FilePath]
applyModifiers fs ms = do
    foldM apply fs ms
  where
    -- Apply modifier to file list
    -- apply :: [FilePath] -> Modifier -> IO [FilePath]
    apply fs m = case m of
        -- Absolute path (sort afterwards to match Zsh behavior)
        "a" -> sortM $ mapM makeAbsolute fs
        -- Resolved path (sort afterwards to match Zsh behavior)
        "A" -> sortM $ mapM canonicalizePath fs
        -- Dirname
        "h" -> return $ map takeDirectory fs
        -- To lowercase
        "l" -> return $ sort (map (map toLower) fs)
        -- Drop ext
        "r" -> return $ map dropExtension fs
        -- Basename
        "t" -> return $ map takeFileName fs
        -- To uppercase
        "u" -> return $ sort (map (map toUpper) fs)
        otherwise -> return fs
    sortM = liftM sort

extractQualifiersAndModifiers :: String -> Maybe (String, [Qualifier], [Modifier])
extractQualifiersAndModifiers s =
    if s == ""
    then Just (s, [], [])
    else if last s == ')'
         then case parseQualifiersAndModifiers suffix of
                  Nothing            -> Nothing
                  Just (quals, mods) -> Just (rest, quals, mods)
         else Just (s, [], [])
  where
    -- reverseExtract :: String -> String -> (String, String)
    reverseExtract s acc =
        if head s == '('
        then (tail s, acc)
        else if (head s) `elem` (")|" :: String)
             -- No qualifiers or modifiers, hand back the original string
             then (")" ++ acc ++ s, "")
             else reverseExtract (tail s) (acc ++ [head s])
    (rrest, rsuffix) = reverseExtract (tail $ reverse s) ""
    rest = reverse rrest
    suffix = reverse rsuffix

parseQualifiersAndModifiers :: String -> Maybe ([Qualifier], [Modifier])
parseQualifiersAndModifiers s =
    case parseOnly qualifiersAndModifiers (pack s) of
        Left _ -> Nothing
        Right qm  -> Just qm

qualifiersAndModifiers :: Parser ([Qualifier], [Modifier])
qualifiersAndModifiers = do
    quals <- many qualifier
    mods <- many modifier
    endOfInput
    return (quals, mods)

-- TODO: Expand the list of qualifiers
qualifier :: Parser Qualifier
qualifier = liftM unpack (choice ["/", ".", "@", "-/", "-.",
                                  "on", "oL", "ol", "oa", "om", "oc"])

-- TODO: Expand the list of modifiers
modifier :: Parser Modifier
modifier = liftM unpack (":" *> choice ["a", "A", "h", "l", "r", "t", "u"])

matchBuilder :: FilePath -> [PatternSegment] -> IO [FilePath]
matchBuilder _ [] = do
    return [""]
matchBuilder root (p:ps) = do
    items <- liftM (\xs -> filter notSpecialDirectory xs)
                   (catch (getDirectoryContents root) getDirectoryContentsFailed)
    -- Filter out files if this is not the last segment yet
    items' <- do
        if ps == []
            then return items
            else filterM doesSubDirectoryExist items
    -- Filter out items matching pattern
    let items'' = filter (\f -> f `matchesSegment` p) items'
    if ps == []
        then return items''
        else liftM concat (mapM recurse items'')
  where
    -- isSpecialDirectory :: FilePath -> Bool
    notSpecialDirectory f = f /= "." && f /= ".."
    -- getDirectoryContentsFailed :: IO [FilePath]
    getDirectoryContentsFailed e = do
        -- TODO: Refactor: Return an error instead of printing to stderr
        hPutStrLn stderr ("Error: Failed to list directory" ++ root ++ ": " ++ show (e :: IOError))
        return []
    -- doesSubDirectoryExist :: FilePath -> IO Bool
    doesSubDirectoryExist f = doesDirectoryExist (root </> f)
    -- recurse :: FilePath -> IO [FilePath]
    recurse sub =  do
        subMatches <- matchBuilder (root </> sub) ps
        return $ map (\f -> sub </> f) subMatches

-- |Special characters that need quoting.
--
-- ']', '<', and ')' are temporarily on the list until we figure out a better
-- solution. '/' is on the list because glob operators (except ** or ***)
-- cannot extend over multiple path segments.
specialChars :: [Char]
specialChars = "*?[]<()|/"

-- |Placeholder for unimplemented parsers.
unimplementedParser :: Parser a
unimplementedParser = do
    fail "unimplemented parser"

-- |Parse out absolute and pattern segments.
pattern :: Parser (Bool, [PatternSegment])
pattern = do
    c <- peekChar'
    let absolute = c == '/'
    when (c == '/') (void $ char '/')
    p <- patternSegment
    ps <- many ((many1 $ char '/') *> patternSegment)
    endOfInput
    return (absolute, (p:ps))

patternSegment :: Parser PatternSegment
patternSegment = many patternUnit

patternUnit :: Parser PatternUnit
patternUnit = choice [literal, star, doubleStar, tripleStar, questionMark,
                      charClass, invertedCharClass, numberRange, group,
                      alternation]

literal :: Parser PatternUnit
literal = liftM Literal (many1 literalChar)
  where
    metachars = specialChars
    literalChar = do
        c <- anyChar
        if c `elem` metachars
           then fail "encountered special character"
           -- If the current char is backslash, the literal char is the next one
           else if c /= '\\'
                then return c
                else do
                    -- TODO: Handle \/
                    c2 <- anyChar
                    return c2

star :: Parser PatternUnit
star = do
    _ <- char '*'
    next <- peekChar
    if next == Just '*'
        then fail "more than one consecutive stars"
        else return Star

doubleStar :: Parser PatternUnit
doubleStar = unimplementedParser

tripleStar :: Parser PatternUnit
tripleStar = unimplementedParser

questionMark :: Parser PatternUnit
questionMark = do
    _ <- char '?'
    return QuestionMark

-- |Helper parser for charClass and invertedCharClass.
charRanges :: Parser [Char]
charRanges = do
    ranges <- many charRange
    return $ concat ranges
  where
    -- charRange :: Parser [Char]
    charRange = do
        c1 <- anyChar
        if c1 == '-'
            then do
            -- - must be the last character in the class
            c2 <- peekChar'
            if c2 == ']'
                then return ['-']
                else fail "dash not at the end of a char class"
            else if c1 == ']'
            then fail "end of char class reached"
            else do
            c2 <- peekChar'
            if c2 /= '-'
                -- Not a range, or in other words, singleton range
                then return [c1]
                -- Either a range, or - is the last character in the class
                else do
                _ <- anyChar -- skip c2
                c3 <- peekChar'
                if c3 == ']'
                    then return [c1, '-']
                    else do
                    _ <- anyChar -- skip c3
                    return $ map chr [ord c1..ord c3]

charClass :: Parser PatternUnit
charClass = do
    _ <- char '['
    first <- peekChar'
    if first `elem` ("^!" :: String)
        then fail "inverted char class"
        else do
        included <- charRanges
        _ <- char ']'
        return $ CharClass included

invertedCharClass :: Parser PatternUnit
invertedCharClass = do
    _ <- char '['
    _ <- choice [char '^', char '!']
    excluded <- charRanges
    _ <- char ']'
    return $ InvertedCharClass excluded

numberRange :: Parser PatternUnit
numberRange = do
    _ <- char '<'
    lowerStr <- many digit
    _ <- char '-'
    upperStr <- many digit
    _ <- char '>'
    let lower = if lowerStr == "" then 0 else read lowerStr
    -- Use -1 to indicate no upper bound
    let upper = if upperStr == "" then -1 else read upperStr
    return $ NumberRange lower upper

group :: Parser PatternUnit
group = do
    _ <- char '('
    seg <- patternSegment
    _ <- char ')'
    return $ Group seg

alternation :: Parser PatternUnit
alternation = do
    _ <- char '('
    seg1 <- patternSegment
    otherSegs <- many1 (char '|' *> patternSegment)
    _ <- char ')'
    return $ Alternation (seg1:otherSegs)

matchesSegment :: String -> PatternSegment -> Bool
matchesSegment s [] = s == ""
matchesSegment s ((Literal ls):xs) =
    case (stripPrefix ls s) of
        Nothing   -> False
        Just rest -> matchesSegment rest xs
matchesSegment s (Star:xs) =
    -- Divide at arbitrary point
    or $ map matchesAtDivisionPoint [0..length s]
  where
    matchesAtDivisionPoint n = matchesSegment suffix xs
      where
        suffix = drop n s
matchesSegment s (DoubleStar:xs) =
    False -- TODO
matchesSegment s (TripleStar:xs) =
    False -- TODO
matchesSegment s (QuestionMark:xs) =
    case s of
        ""      -> False
        _:rest -> matchesSegment rest xs
matchesSegment s ((CharClass cs):xs) =
    case s of
        ""     -> False
        c:rest -> (c `elem` cs) && (matchesSegment rest xs)
matchesSegment s ((InvertedCharClass cs):xs) =
    case s of
        ""     -> False
        c:rest -> (c `notElem` cs) && (matchesSegment rest xs)
matchesSegment s ((NumberRange l u):xs) =
    or $ map matchesAtDivisionPoint [1..firstNonDigitIndex]
  where
    -- Calculate first non-digit index
    nd ss acc = if ss == ""
                then acc
                else if Data.Char.isDigit (head ss)
                then nd (tail ss) (acc + 1)
                else acc
    firstNonDigitIndex = nd s 0
    matchesAtDivisionPoint n =
        (inRange $ read prefix) && (matchesSegment suffix xs)
      where
        prefix = Prelude.take n s
        suffix = drop n s
        inRange num = (l <= num) && ((num <= u) || (u < 0))
matchesSegment s ((Group seg):xs) =
    -- Divide at arbitrary point
    or $ map matchesAtDivisionPoint [0..length s]
  where
    matchesAtDivisionPoint n =
        (matchesSegment prefix seg) && (matchesSegment suffix xs)
      where
        prefix = Prelude.take n s
        suffix = drop n s
matchesSegment s ((Alternation segs):xs) =
    -- Divide at arbitrary point
    or $ map matchesAtDivisionPoint [0..length s]
  where
    matchesAtDivisionPoint n =
        (or $ map (\seg -> matchesSegment prefix seg) segs) &&
        (matchesSegment suffix xs)
      where
        prefix = Prelude.take n s
        suffix = drop n s
